unit Crypto;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

uses SysUtils, DECHash, DECUtil;

procedure Hmac(Hash: TDECHashClass; const Text; TextLen: Cardinal; Key: PByte; KeyLen: Integer; var Digest: RawByteString);
function PBKDF2(const Data; DataLen: Cardinal; const Salt; SaltLen: Cardinal; Res: PByte; ResLen, Rounds: Integer): Boolean;
procedure RemovePKCS5Padding(var B: TBytes);

implementation

procedure Hmac(Hash: TDECHashClass; const Text; TextLen: Cardinal; Key: PByte; KeyLen: Integer; var Digest: RawByteString);
var
  Ctx: TDECHash;
  TK: RawByteString;
  KPad: array of Byte;
  i: Integer;
begin
  SetLength(KPad, Hash.BlockSize);
  if KeyLen > Hash.BlockSize then
  begin
    TK := Hash.CalcBuffer(Key^, KeyLen);

    Key := @TK[1];
    KeyLen := Hash.DigestSize;
  end;

  FillChar(KPad[0], Length(KPad), 0);
  Move(Key^, KPad[0], KeyLen);
  for i := 0 to High(KPad) do
    KPad[i] := KPad[i] xor $36;

  Ctx := Hash.Create;
  try
    Ctx.Init;
    Ctx.Calc(KPad[0], Hash.BlockSize);
    Ctx.Calc(Text, TextLen);
    Ctx.Done;
    Digest := Ctx.DigestStr;
  finally
    Ctx.Free;
  end;

  FillChar(KPad[0], Length(KPad), 0);
  Move(Key^, KPad[0], KeyLen);
  for i := 0 to High(KPad) do
    KPad[i] := KPad[i] xor $5C;

  Ctx := Hash.Create;
  try
    Ctx.Init;
    Ctx.Calc(KPad[0], Hash.BlockSize);
    Ctx.Calc(Digest[1], Hash.DigestSize);
    Ctx.Done;
    Digest := Ctx.DigestStr;
  finally
    Ctx.Free;
  end;
end;


function bswap32(V: Cardinal): Cardinal;
asm
  bswap eax // Switch endianness
end;

function PBKDF2(const Data; DataLen: Cardinal; const Salt; SaltLen: Cardinal; Res: PByte; ResLen, Rounds: Integer): Boolean;
var
  ASalt: array of Byte;
  Count, i, j, R: Cardinal;
  D1, D2, OBuf: RawByteString;
begin
  if (Rounds < 1) or (ResLen < 1) then
    Exit(False);

  SetLength(ASalt, SaltLen + 4);
  Move(Salt, ASalt[0], SaltLen);

  SetLength(OBuf, THash_SHA1.DigestSize);
  Count := 1;
  while ResLen > 0 do
  begin
    PCardinal(@ASalt[SaltLen])^ := bswap32(Count);
    Hmac(THash_SHA1, ASalt[0], SaltLen + 4, @Data, DataLen, D1);
    Move(D1[1], OBuf[1], Length(OBuf));

    for i := 2 to Rounds do
    begin
      Hmac(THash_SHA1, D1[1], Length(D1), @Data, DataLen, D2);
      Move(D2[1], D1[1], Length(D1));
      for j := 1 to Length(OBuf) do
        OBuf[j] := AnsiChar(Byte(OBuf[j]) xor Byte(D1[j]));
    end;

    if ResLen < THash_SHA1.DigestSize then
      R := ResLen
    else
      R := THash_SHA1.DigestSize;

    Move(OBuf[1], Res^, R);
    Inc(Res, R);
    Dec(ResLen, R);

    Inc(Count);
  end;

  FillChar(ASalt[0], Length(ASalt), 0);
  ProtectBinary(D1);
  ProtectBinary(D2);
  ProtectBinary(OBuf);
  Result := True;
end;

procedure RemovePKCS5Padding(var B: TBytes);
var
  Pad: Byte;
  i: Integer;
begin
  Pad := B[High(B)];
  if (Pad > 0) and (Pad <= 16) then
  begin
    for i := High(B) - 1 downto Length(B) - Pad do
      if B[i] <> Pad then
      begin
        Pad := 0;
        Break;
      end;
    SetLength(B, Length(B) - Pad);
  end;
end;

end.
