unit Eratosthenes;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

uses Classes;

type
  TSieveOfEratosthenes = class
  private
    FMaxValue: Cardinal;
    FSieve: TBits;
    FPrimes: array of Cardinal;
    FPrimesFound: Cardinal;
    function GetPrimes(i: Cardinal): Cardinal;
    function GetPrimesFound: Cardinal;
    procedure FindPrimes;
  public
    constructor Create(AMaxValue: Cardinal);
    destructor Destroy; override;

    property Primes[Index: Cardinal]: Cardinal read GetPrimes;
    property PrimesFound: Cardinal read GetPrimesFound;
  end;

implementation

{ TSieveOfEratosthenes }

constructor TSieveOfEratosthenes.Create(AMaxValue: Cardinal);
begin
  FSieve := TBits.Create;
  FMaxValue := AMaxValue;
  FindPrimes;
end;

destructor TSieveOfEratosthenes.Destroy;
begin
  FSieve.Free;
  inherited;
end;

procedure TSieveOfEratosthenes.FindPrimes;
var
  i, j: Cardinal;
begin
  FSieve.Size := FMaxValue + 1;

  i := 2;
  while i * i <= FMaxValue do
  begin
    if FSieve.Bits[i] = false then
    begin
      j := i * i;
      while j <= FMaxValue do
      begin
        FSieve.Bits[j] := true;
        Inc(j, i);
      end;
    end;
    Inc(i);
  end;

  FPrimesFound := 0;
  for i := 2 to FMaxValue do
    if not FSieve.Bits[i] then
      Inc(FPrimesFound);

  SetLength(FPrimes, FPrimesFound);
  j := 0;
  for i := 2 to FMaxValue do
  begin
    if not FSieve.Bits[i] then
    begin
      FPrimes[j] := i;
      Inc(j);
    end;
  end;
end;

function TSieveOfEratosthenes.GetPrimes(i: Cardinal): Cardinal;
begin
  Result := FPrimes[i];
end;

function TSieveOfEratosthenes.GetPrimesFound: Cardinal;
begin
  Result := Length(FPrimes);
end;

end.
