unit IonSymbols;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

uses SysUtils, Classes;

const
  UNKNOWN_SYMBOL_ID = -1;
  SID_ION = 1;
  SID_ION_1_0 = 2;
  SID_ION_SYMBOL_TABLE = 3;
  SID_NAME = 4;
  SID_VERSION = 5;
  SID_IMPORTS = 6;
  SID_SYMBOLS = 7;
  SID_MAX_ID = 8;
  SID_ION_SHARED_SYMBOL_TABLE = 9;
  ION_1_0_MAX_ID = 10;

type
  TSystemSymbols = class
  public const
    ION = '$ion';
    ION_1_0 = '$ion_1_0';
    ION_SYMBOL_TABLE = '$ion_symbol_table';
    NAME = 'name';
    VERSION = 'version';
    IMPORTS = 'imports';
    SYMBOLS = 'symbols';
    MAX_ID = 'max_id';
    ION_SHARED_SYMBOL_TABLE = '$ion_shared_symbol_table';
  end;

  TIonCatalogItem = class
  public
    Name: string;
    Version: Integer;
    SymNames: TStringList;
    constructor Create(const AName: string; AVersion: Integer; ASymNames: TStringList);
  end;

  TSymbolToken = record
    Text: string;
    SID: Integer;
    constructor Create(const AText: string; ASID: Integer);
  end;

  TSymbolTable = class
  private
    FTable: array of string;
  public
    constructor Create;

    function FindByID(ID: Integer): string;
    procedure Import(Table: TIonCatalogItem; MaxID: Integer);
    procedure ImportUnknown(const Name: string; MaxID: Integer);
  end;

implementation

{ TSymbolToken }

constructor TSymbolToken.Create(const AText: string; ASID: Integer);
begin
  if (Text = '') and (ASID = 0) then
    raise EArgumentException.Create('Symbol token must have Text or SID');

  Text := AText;
  SID := ASID;
end;

{ TSymbolTable }

constructor TSymbolTable.Create;
begin
  SetLength(FTable, 10);
  FTable[SID_ION] := TSystemSymbols.ION;
  FTable[SID_ION_1_0] := TSystemSymbols.ION_1_0;
  FTable[SID_ION_SYMBOL_TABLE] := TSystemSymbols.ION_SYMBOL_TABLE;
  FTable[SID_NAME] := TSystemSymbols.NAME;
  FTable[SID_VERSION] := TSystemSymbols.VERSION;
  FTable[SID_IMPORTS] := TSystemSymbols.IMPORTS;
  FTable[SID_SYMBOLS] := TSystemSymbols.SYMBOLS;
  FTable[SID_MAX_ID] := TSystemSymbols.MAX_ID;
  FTable[SID_ION_SHARED_SYMBOL_TABLE] := TSystemSymbols.ION_SHARED_SYMBOL_TABLE;
end;

function TSymbolTable.FindByID(ID: Integer): string;
begin
  if ID < 1 then
    raise EArgumentException.Create('Invalid symbol id');

  if ID < Length(FTable) then
    Result := FTable[ID]
  else
    Result := '';
end;

procedure TSymbolTable.Import(Table: TIonCatalogItem; MaxID: Integer);
var
  i, Old: Integer;
begin
  Old := Length(FTable);
  SetLength(FTable, Old + MaxID);
  for i := 0 to MaxID - 1 do
  begin
    FTable[Old + i] := Table.SymNames[i];
  end;
end;

procedure TSymbolTable.ImportUnknown(const Name: string; MaxID: Integer);
var
  i, Old: Integer;
begin
  Old := Length(FTable);
  SetLength(FTable, Old + MaxID);
  for i := 0 to MaxID - 1 do
  begin
    FTable[Old + i] := Name + '#' + IntToStr(i + 1);
  end;
end;

{ TIonCatalogItem }

constructor TIonCatalogItem.Create(const AName: string; AVersion: Integer;
  ASymNames: TStringList);
begin
  Name := AName;
  Version := AVersion;
  SymNames := ASymNames;
end;

end.
