unit Main;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics,
  Controls, Forms, Dialogs, StdCtrls, Generics.Collections,
  BinaryIon, DrmIon, Kinf, KRFTable, IonSymbols, ComCtrls;

type
  TForm1 = class(TForm)
    btnDecrypt: TButton;
    btnPrintVoucher: TButton;
    btnPrintKinf: TButton;
    btnLoadCont: TButton;
    pcMain: TPageControl;
    tsLog: TTabSheet;
    Memo1: TMemo;
    tsKFX: TTabSheet;
    lblCont: TLabel;
    LVEntries: TListView;
    OpenDialog1: TOpenDialog;
    btnPrintDRMION: TButton;
    SaveDialog1: TSaveDialog;
    procedure btnDecryptClick(Sender: TObject);
    procedure btnPrintVoucherClick(Sender: TObject);
    procedure btnPrintKinfClick(Sender: TObject);
    procedure btnLoadContClick(Sender: TObject);
    procedure LVEntriesDblClick(Sender: TObject);
    procedure btnPrintDRMIONClick(Sender: TObject);
  private
    FBookPath: string;

    procedure LoadCont(const FN: string);
    function OnVoucherRequired(const Name: string): TDrmIonVoucher;
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

function TForm1.OnVoucherRequired(const Name: string): TDrmIonVoucher;
var
  FS: TFileStream;
begin
  FS := TFileStream.Create(FBookPath + Name + '.voucher', fmOpenRead or fmShareDenyNone);
  Result := TDrmIonVoucher.Create(FS);
end;

procedure DRMIONPrepare(MS: TMemoryStream);
const
  MAGIC: AnsiString = #$EA'DRMION'#$EE;
var
  S: AnsiString;
begin
  // First and last 8 bytes are not part of ION spec
  MS.SetSize(MS.Size - 8);
  SetLength(S, 8);
  MS.Read(S[1], 8);
  if S <> MAGIC then
    raise Exception.Create('DRMION header not found in file');
end;

procedure TForm1.btnDecryptClick(Sender: TObject);
var
  MS: TMemoryStream;
  Ion: TDrmIon;
begin
  if not OpenDialog1.Execute then
    Exit;

  MS := TMemoryStream.Create;
  try
    FBookPath := ExtractFilePath(OpenDialog1.FileName);
    MS.LoadFromFile(OpenDialog1.FileName);
    DRMIONPrepare(MS);
    Ion := TDrmIon.Create(MS, OnVoucherRequired);
    if SaveDialog1.Execute then
      Ion.Parse(SaveDialog1.FileName);
    Ion.Free;
  finally
    MS.Free;
  end;
end;

procedure TForm1.btnPrintDRMIONClick(Sender: TObject);
var
  MS: TMemoryStream;
  Ion: TDrmIon;
begin
  if not OpenDialog1.Execute then
    Exit;

  MS := TMemoryStream.Create;
  try
    FBookPath := ExtractFilePath(OpenDialog1.FileName);
    MS.LoadFromFile(OpenDialog1.FileName);
    DRMIONPrepare(MS);
    Ion := TDrmIon.Create(MS, OnVoucherRequired);
    Memo1.Lines.BeginUpdate;
    try
      Ion.Print(Memo1.Lines);
    finally
      Memo1.Lines.EndUpdate;
    end;
    Ion.Free;
  finally
    MS.Free;
  end;
end;

procedure TForm1.btnPrintVoucherClick(Sender: TObject);
var
  FS: TFileStream;
  Ion: TDrmIonVoucher;
begin
  if not OpenDialog1.Execute then
    Exit;

  FS := TFileStream.Create(OpenDialog1.FileName, fmOpenRead or fmShareDenyNone);
  try
    Ion := TDrmIonVoucher.Create(FS);
    Ion.Parse;
    Ion.PrintEnvelope(Memo1.Lines);
    Ion.PrintVoucher(Memo1.Lines);
    Ion.PrintKey(Memo1.Lines);
    Ion.Free;
  finally
    FS.Free;
  end;
end;

procedure TForm1.btnPrintKinfClick(Sender: TObject);
var
  Kinf: TKinfStorage;
  E: TPair<string, string>;
begin
  Kinf := ReadKinf(GetKinfPath);
  for E in Kinf do
    Memo1.Lines.Add(E.Key + ':' + E.Value);
end;

type
  TEntity = packed record
    ID, TypeIndex: Integer;
    Offset, Length: Int64;
  end;

procedure TForm1.btnLoadContClick(Sender: TObject);
begin
  if OpenDialog1.Execute then
    LoadCont(OpenDialog1.FileName);
end;

procedure TForm1.LoadCont(const FN: string);
var
  FS: TFileStream;
  MS: TMemoryStream;
  Ion: TBinaryIonParser;
  Magic, HeaderLen, InfoOffset, InfoLen, IndexOffset, IndexLen: Cardinal;
  Version: Word;
  Ent: TEntity;
  OldPos: Int64;
begin
  FS := TFileStream.Create(FN, fmOpenRead or fmShareDenyNone);
  try
    FS.Read(Magic, 4);
    if Magic <> $544E4F43 then
      raise Exception.Create('File does not start with a CONT header');

    FS.Read(Version, 2);
    FS.Read(HeaderLen, 4);
    FS.Read(InfoOffset, 4);
    FS.Read(InfoLen, 4);

    lblCont.Caption := Format('CONT v%d. InfoOffset: %d, InfoLen: %d.', [Version, InfoOffset, InfoLen]);

    IndexOffset := 0;
    IndexLen := 0;

    FS.Seek(InfoOffset, soBeginning);
    MS := TMemoryStream.Create;
    try
      MS.CopyFrom(FS, InfoLen);
      MS.Position := 0;
      Ion := TBinaryIonParser.Create(MS);
      Ion.ForceImport(GetKRFTable);

      Ion.Next;
      Ion.StepIn;
      while Ion.HasNext do
      begin
        Ion.Next;
        if Ion.GetFieldName = 'bcIndexTabOffset' then
          IndexOffset := Ion.IntValue
        else if Ion.GetFieldName = 'bcIndexTabLength' then
          IndexLen := Ion.IntValue;
      end;
      Ion.StepOut;

      Ion.Print(Memo1.Lines);
      Ion.Free;
    finally
      MS.Free;
    end;

    FS.Seek(IndexOffset, soBeginning);
    while FS.Position < IndexLen + IndexOffset do
    begin
      FS.Read(Ent, SizeOf(Ent));
      with LVEntries.Items.Add do
      begin
        Caption := IntToStr(Ent.ID);
        SubItems.Add(KRF_TABLE[Ent.TypeIndex - ION_1_0_MAX_ID]);
        SubItems.Add(IntToStr(Ent.Offset));
        SubItems.Add(IntToStr(Ent.Length));
        OldPos := FS.Position;
        FS.Seek(Ent.Offset + HeaderLen, soBeginning);
        MS := TMemoryStream.Create;
        MS.CopyFrom(FS, Ent.Length);
        MS.Position := 10;
        Data := MS;
        FS.Position := OldPos;
      end;
    end;
  finally
    FS.Free;
  end;
end;

procedure TForm1.LVEntriesDblClick(Sender: TObject);
begin
  if LVEntries.Selected = nil then
    Exit;

  with TBinaryIonParser.Create(TStream(LVEntries.Selected.Data)) do
  begin
    ForceImport(GetKRFTable);
    Print(Memo1.Lines);
    Free;
  end;
  pcMain.TabIndex := 0;
end;

end.
